import os
import pytest

import green_ac
import numpy as np


def test_two_peaks():
    imgrid = (2 * np.linspace(-50, 49, 100) + 1) * np.pi / 2
    grid = np.linspace(-2, 2, 1001)
    eta = 0.01
    data = 0.5 * (1 / (1.j * imgrid + 0.5) + 1 / (1.j * imgrid - 0.5))
    data_test = 0.5 * (1 / (grid + eta * 1.j + 0.5) + 1 / (grid + eta * 1.j - 0.5))

    with pytest.raises(Exception) as e:
        green_ac.solve("Nevanlina", imgrid, grid, data, eta=eta)
    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data, eta=eta)

    assert np.allclose(data_test, data_out)


def test_default_eta():
    imgrid = (2 * np.linspace(-50, 49, 100) + 1) * np.pi / 2
    grid = np.linspace(-2, 2, 1001)
    eta = 0.01
    data = 0.5 * (1 / (1.j * imgrid + 0.5) + 1 / (1.j * imgrid - 0.5))
    data_test = 0.5 * (1 / (grid + eta * 1.j + 0.5) + 1 / (grid + eta * 1.j - 0.5))

    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data)

    assert np.allclose(data_test, data_out)


def test_two_orbitals():
    imgrid = (2 * np.linspace(0, 100, 101) + 1) * np.pi / 2
    grid = np.linspace(-2, 2, 1001)
    eta = 0.01
    data = np.zeros(imgrid.shape + (2,), dtype=np.complex128)
    data[:, 0] = 0.5 * (1 / (1.j * imgrid + 0.5) + 1 / (1.j * imgrid - 0.5))
    data[:, 1] = 0.5 * (1 / (1.j * imgrid + 0.2) + 1 / (1.j * imgrid - 0.2))
    data_test_1 = 0.5 * (1 / (grid + eta * 1.j + 0.5) + 1 / (grid + eta * 1.j - 0.5))
    data_test_2 = 0.5 * (1 / (grid + eta * 1.j + 0.2) + 1 / (grid + eta * 1.j - 0.2))

    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data, eta=eta)

    assert data_out.shape == (grid.shape[0], 2)

    assert np.allclose(data_test_1, data_out[:, 0])
    assert np.allclose(data_test_2, data_out[:, 1])


def test_two_orbitals_momenta():
    imgrid = (2 * np.linspace(0, 100, 101) + 1) * np.pi / 2
    grid = np.linspace(-2, 2, 1001)
    eta = 0.01
    data = np.zeros(imgrid.shape + (10, 2,), dtype=np.complex128)
    for k in range(10):
        data[:, k, 0] = 0.5 * (1 / (1.j * imgrid + 0.5) + 1 / (1.j * imgrid - 0.5))
        data[:, k, 1] = 0.5 * (1 / (1.j * imgrid + 0.2) + 1 / (1.j * imgrid - 0.2))
    data_test_1 = 0.5 * (1 / (grid + eta * 1.j + 0.5) + 1 / (grid + eta * 1.j - 0.5))
    data_test_2 = 0.5 * (1 / (grid + eta * 1.j + 0.2) + 1 / (grid + eta * 1.j - 0.2))

    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data, eta=eta)

    assert data_out.shape == (grid.shape[0], 10, 2)

    assert np.allclose(data_test_1, data_out[:, 0, 0])
    assert np.allclose(data_test_2, data_out[:, 1, 1])


def test_slicing():
    imgrid = (2 * np.linspace(0, 100, 101) + 1) * np.pi / 2
    grid = np.linspace(-2, 2, 1001)
    eta = 0.01
    N = 10
    data = np.zeros(imgrid.shape + (N, 2,), dtype=np.complex128)
    data_test = np.zeros(grid.shape + (N, 2,), dtype=np.complex128)
    eps_k = np.cos(np.linspace(0, 2 * np.pi, N, endpoint=False))
    for k in range(10):
        data[:, k, 0] = 0.5 * (1 / (1.j * imgrid + 0.5 - eps_k[k]) + 1 / (1.j * imgrid - 0.5 - eps_k[k]))
        data[:, k, 1] = 0.5 * (1 / (1.j * imgrid + 0.2 - eps_k[k]) + 1 / (1.j * imgrid - 0.2 - eps_k[k]))
        data_test[:, k, 0] = 0.5 * (1 / (grid + eta * 1.j + 0.5 - eps_k[k]) + 1 / (grid + eta * 1.j - 0.5 - eps_k[k]))
        data_test[:, k, 1] = 0.5 * (1 / (grid + eta * 1.j + 0.2 - eps_k[k]) + 1 / (grid + eta * 1.j - 0.2 - eps_k[k]))
    data2 = np.copy(data)

    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data[:, 2, :], eta=eta)
    assert data_out.shape == (grid.shape[0], 2)

    assert np.allclose(data_test[:, 2, :], data_out[:, :])
    assert np.allclose(data, data2)

    data_out = green_ac.solve("Nevanlinna", imgrid, grid, data[:, 2:4, :], eta=eta)

    assert data_out.shape == (grid.shape[0], 2, 2)

    assert np.allclose(data_test[:, 3, 0], data_out[:, 1, 0])
    assert np.allclose(data_test[:, 2, 1], data_out[:, 0, 1])
    assert np.allclose(data, data2)


def test_from_file():
    data_dir = os.path.abspath(os.path.dirname(__file__)) + '/data'
    Xiw = np.loadtxt(data_dir + "/X_iw.txt")
    Xw = np.loadtxt(data_dir + "/X_w.txt")
    mgrid = Xiw[:, 0]
    grid = Xw[:, 0]
    data = Xiw[:, 1] + 1.j * Xiw[:, 2]
    test_data = Xw[:, 1]
    data_out = green_ac.solve("Nevanlinna", mgrid, grid, data, eta=0.01)
    np.allclose(test_data, -data_out.imag / np.pi, atol=1e-3)



import concurrent.futures
import functools
import importlib.util as ilu
import multiprocessing
import os

import numpy as np


def default_kwargs(**defaultKwargs):
    def actual_decorator(fn):
        @functools.wraps(fn)
        def g(*args, **kwargs):
            defaultKwargs.update(kwargs)
            return fn(*args, **defaultKwargs)

        return g

    return actual_decorator


has_mpi = False

if ilu.find_spec("mpi4py") is not None:
    from mpi4py import MPI

    has_mpi = True

from ._green_ac import solve as _solve


def _solve_nevanlinna_conc(kind, mgrid, grid, data_rs, prec):
    """
    This is a wrapper around C++ function to make it work with ProcessPoolExecutor
    """
    return _solve(kind, mgrid, grid, data_rs, prec)


def _solve_nevanlinna_mpi():
    pass


@default_kwargs(eta=0.01, prec=128)
def _solve_nevanlinna(mgrid, grid, data, **kwargs):
    eta = kwargs["eta"]
    prec = kwargs["prec"]
    from multiprocessing import cpu_count
    # Find the number of cpus to use for parallelization of analytic continuation
    slurm_env_var = 'SLURM_JOB_CPUS_PER_NODE'
    phys_ncpu = cpu_count()
    _ncpus = int(os.environ[slurm_env_var]) if slurm_env_var in os.environ else int(phys_ncpu // 2)
    # linearize data for easy chunking
    lin_s = data.size // data.shape[0]
    data_rs = data.reshape([data.shape[0], lin_s])
    out_data_rs = np.zeros([grid.shape[0], lin_s], dtype=np.complex128)
    fs = []
    chunks = np.array_split(range(lin_s), _ncpus)
    # create a process pool
    with concurrent.futures.ProcessPoolExecutor(_ncpus, mp_context=multiprocessing.get_context('forkserver')) as executor:
        for i in chunks:
            if len(i) == 0:
                continue
            f = executor.submit(_solve_nevanlinna_conc, "Nevanlinna", mgrid * 1.j, grid + eta * 1.j, data_rs[:, i],
                                prec)
            fs.append(f)
    # wait for completion
    concurrent.futures.wait(fs)
    # get data from each process
    i = 0
    for c in (chunks):
        if len(c) == 0:
            continue
        out_data_rs[:, c] = fs[i].result()
        i += 1
    out_data = out_data_rs.reshape(grid.shape + data.shape[1:])
    return out_data


solvers = {"NEVANLINNA": _solve_nevanlinna}
solvers_mpi = {"NEVANLINNA": _solve_nevanlinna_mpi}


@default_kwargs(eta=0.01, prec=128)
def solve_mpi(comm, kind, mgrid, grid, data, **kwargs):
    """
    Solve analytical continuation problem using MPI parallelization
    MPI4Py should be available and properly built

    :param comm: MPI communicator
    :param kind: type of continuation (should be string)
    :param mgrid: Matsubara frequency points (real one dimensional numpy array)
    :param grid: real frequency points (real one dimensional numpy array)
    :param data: Matsubara frequency Green's function (complex N-dimensional numpy array)
    :param kwargs: additional solver specific parameters such as broadening `eta` or precision
    :return: complex N-dimensional numpy array with real frequency Green's function
    """
    if not has_mpi:
        raise RuntimeError("MPI is not found")
    prec = kwargs["prec"]
    eta = kwargs["eta"]
    lin_s = data.size // data.shape[0]
    data_rs = data.reshape([data.shape[0], lin_s])
    out_data_rs = np.zeros([grid.shape[0], lin_s], dtype=np.complex128)
    rank = comm.Get_rank()
    size = comm.Get_size()
    for i in range(rank, lin_s, size):
        out_data_rs[:, i] = _solve(kind, mgrid * 1.j, grid + eta * 1.j, data_rs[:, i], prec)
    comm.Allreduce(MPI.IN_PLACE, out_data_rs.view(np.float64), op=MPI.SUM)
    out_data = out_data_rs.reshape(grid.shape + data.shape[1:])
    return out_data


def solve(kind, mgrid, grid, data, **kwargs):
    """Solve analytical continuation using Python built-in concurrent parallelism

    :param kind: type of continuation (should be string)
    :param mgrid: Matsubara frequency points (real one dimensional numpy array)
    :param grid: real frequency points (real one dimensional numpy array)
    :param data: Matsubara frequency Green's function (complex N-dimensional numpy array)
    :param kwargs: additional solver specific parameters such as broadening `eta` or precision
    :return: complex N-dimensional numpy array with real frequency Green's function
    """
    if not kind.upper() in solvers.keys():
        raise RuntimeError(f"Unknown continuation kind {kind}, currently only {solvers.keys()} are supported.")
    return solvers[kind.upper()](mgrid, grid, data, **kwargs)
